/* devAvme9440.c */

/***********************************************************************
* Copyright (c) 2007 UChicago Argonne LLC, as Operator of Argonne
* National Laboratory, and the Regents of the University of
* California, as Operator of Los Alamos National Laboratory.
* acromag is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
***********************************************************************/

/*****************************************************************
 *
 *      Author :                     Greg Nawrocki
 *      Date:                        4/93
 *
 * Modification Log:
 * -----------------
 *      ...
 */

/*To Use this device, Include the following before iocInit */
/* devAvme9440Config(ncards,a16base,intvecbase)		*/
/* For Example						*/
/* devAvme9440Config(1,0x2800,0x78)			*/

 /*
  * This device support routine provides EPICS support for the      *
  * Acromag AVME-9440 16 bit binary input and output board.  Change *
  * of state I/O interrupts are available for binary input signals  *
  * 0 -7 only, and only for the BI record.  If interrupts are       *
  * desired for MBBI type values, they must be done via BI records  *
  * linked to CALC records.  As many I/O interrupt scan BI records  *
  * may be connected to a single binary input signal as desired.    *
  *
  */

 /*********************************************************************/
 /**							             **/
 /**  Binary Outputs 0 - 15 ========== "AVME9440 O" Signals 0 - 15   **/
 /**  Binary Input 0 - 15 ============ "AVME9440 I" Signals 0 - 15   **/
 /**  Binary Output Readback 0 - 15 == "AVME9440 I" Signals 16 - 31  **/
 /**                                                                 **/
 /*********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "alarm.h"
#include "dbDefs.h"
#include "dbScan.h"
#include "dbAccess.h"
#include "devLib.h"
#include "epicsExit.h"
#include "recSup.h"
#include "recGbl.h"
#include "devSup.h"
#include "link.h"
#include "epicsMutex.h"
#include "iocsh.h"

#include "boRecord.h"
#include "biRecord.h"
#include "mbboRecord.h"
#include "mbbiRecord.h"
#include "epicsExport.h"


int  devAvme9440Debug = 0; 
epicsExportAddress(int, devAvme9440Debug);

/***** devAvme9440Debug *****/

/** devAvme9440Debug == 0 --- no debugging messages **/
/** devAvme9440Debug >= 5 --- hardware initialization information **/
/** devAvme9440Debug >= 10 -- record initialization information **/
/** devAvme9440Debug >= 15 -- write commands **/
/** devAvme9440Debug >= 20 -- read commands **/

#define	INPUT_REG	1
#define OUTPUT_REG	2

#define NUM_INT_CHAN 8

struct avme9440 {
  unsigned char pad0[0x80];
  unsigned char pad1;
  unsigned char boardStatus;
  unsigned char pad2[0x1e];
  struct {
    unsigned char pad;
    unsigned char vector;
  } iv[NUM_INT_CHAN];
  unsigned char padb[0x10];
  unsigned char padc;
  unsigned char intStatus;
  unsigned char padd;
  unsigned char intEnable;
  unsigned char pade;
  unsigned char intPolarity;
  unsigned char padf;
  unsigned char intTypeSelect;
  unsigned char padg;
  unsigned char intPaternEnable;
  unsigned short inputData;
  unsigned short outputData;
  unsigned char padh[0x332];
};

struct ioCard {
  volatile struct avme9440 *card;    /* Pointer to this card's registers */
  epicsMutexId lock; 	             /* semaphore */
  int intCnt[NUM_INT_CHAN];  /* # I/O Intr records attached to each channel */
  IOSCANPVT ioscanpvt[NUM_INT_CHAN]; /* interrupt */
};

#define MAX_9440_CARDS 6

static unsigned short BASEADD;
#define LED_INIT        0x02
#define LED_OKRUN       0x03
#define LED_OKINTS      0x0B

#define INT_LEVEL 5
static unsigned short INT_VEC_BASE;

int int_level = INT_LEVEL;

int avme9440_num_links=0;

static struct ioCard cards[MAX_9440_CARDS];
static int init_flag = 0;


static long devAvme9440Report();
static long init(int after);
static void avme9440_isr(void *parm);
static long init_bo_record(struct dbCommon *pbo);
static long init_bi_record(struct dbCommon *pbi);
static long init_mbbo_record(struct dbCommon *pmbbo);
static long init_mbbi_record(struct dbCommon *pmbbi);
static long write_bo(struct boRecord *pbo);
static long read_bi(struct biRecord *pbi);
static long write_mbbo(struct mbboRecord *pmbbo);
static long read_mbbi(struct mbbiRecord *pmbbi);
static long read_card(int card, epicsUInt32 mask, epicsUInt16 *value, int reg);
static long write_card(int card, epicsUInt32 mask, epicsUInt32 value);
static long write_mbbo_card(int card, epicsUInt32 mask, epicsUInt32 value);
static int checkLink(int card);
static long get_bi_int_info(int cmd, struct dbCommon *pbi, IOSCANPVT *ppvt);


/* Create the dset for devBoAvme9440 */
bodset devBoAvme9440 = {
    {
        5,
        devAvme9440Report,
        init,
        init_bo_record,
        NULL,
    },
    write_bo
};
epicsExportAddress(dset,devBoAvme9440);

/* Create the dset for devBiAvme9440 */
bidset devBiAvme9440 = {
    {
        5,
        NULL,
        NULL,
        init_bi_record,
        get_bi_int_info
    },
    read_bi
};
epicsExportAddress(dset,devBiAvme9440);

/* Create the dset for devMbboAvme9440 */
mbbodset devMbboAvme9440 = {
    {
        5,
        NULL,
        NULL,
        init_mbbo_record,
        NULL
    },
    write_mbbo
};
epicsExportAddress(dset,devMbboAvme9440);

/* Create the dset for devMbbiAvme9440 */
mbbidset devMbbiAvme9440 = {
    {
        5,
        NULL,
        NULL,
        init_mbbi_record,
        NULL
    },
    read_mbbi
};
epicsExportAddress(dset,devMbbiAvme9440);

/**************************************************************************
 *
 * Ultra groovy and useful reporting function called from 'dbior'.
 *
 **************************************************************************/
static long devAvme9440Report()
{
    int LinkNum = 0;
    int CardBase = BASEADD;
    int IntVec = INT_VEC_BASE;

    while (LinkNum < avme9440_num_links)
    {
        if (cards[LinkNum].card != NULL)
        {
            printf("    Link %2.2d at 0x%4.4X, IRQ 0x%2.2X, input 0x%4.4X, output 0x%4.4X\n", 
                LinkNum, CardBase, IntVec, 
                cards[LinkNum].card->inputData, 
                cards[LinkNum].card->outputData);

        }
        LinkNum++;
        CardBase += sizeof(struct avme9440);
        IntVec++;
    }
    return 0;
}

/**************************************************************************
*
* Configuration of Driver - how many cards and where?
*
***************************************************************************/
int devAvme9440Config(int ncards, int a16base, int intvecbase)
{
    if (ncards > MAX_9440_CARDS)
    {
        printf("Can't drive more than %d AVME9440 cards\n", MAX_9440_CARDS);
        return -1;
    }
    if (a16base < 0 || a16base > 0x10000 - ncards * sizeof(struct avme9440))
    {
        printf("Bad VMEA16 Base address for %d cards\n", ncards);
        return -1;
    }
    if (intvecbase < 0 || intvecbase > 0x100 - ncards)
    {
        printf("Bad interrupt vector base for %d cards\n", ncards);
        return -1;
    }
    avme9440_num_links = ncards;
    BASEADD = a16base;
    INT_VEC_BASE = intvecbase;
    init(0);
    return 0;
}

static const iocshArg configArg0 = {"ncards", iocshArgInt};
static const iocshArg configArg1 = {"a16base", iocshArgInt};
static const iocshArg configArg2 = {"intvecbase", iocshArgInt};
static const iocshArg *configArgs[] = {
    &configArg0, &configArg1, &configArg2
};
static const iocshFuncDef configFuncDef = {
    "devAvme9440Config", 3, configArgs
};
static void devAvme9440ConfigCallFunc(const iocshArgBuf *args)
{
    devAvme9440Config(args[0].ival, args[1].ival, args[2].ival);
}

static void devAvme9440Registrar()
{
    iocshRegister(&configFuncDef, devAvme9440ConfigCallFunc);
}
epicsExportRegistrar(devAvme9440Registrar);


static void shutdown(void *parm) {
    volatile struct avme9440 *p = (struct avme9440 *) parm;
    p->boardStatus = LED_OKRUN;    /* Disable interrupts */
}

/**************************************************************************
*
* Initialization of AVME9440 Binary I/O Card
*
***************************************************************************/
static long init(int after)
{
  int cardNum, chanNum;
  unsigned char probeVal;
  volatile struct avme9440 *p;

  if (init_flag != 0)
    return 0;

  init_flag = 1;

  if (devRegisterAddress("devAvme9440", atVMEA16, BASEADD,
          sizeof(struct avme9440) * avme9440_num_links, (volatile void **)&p))
  {
    if (devAvme9440Debug >= 5)
       printf("devAvme9440: Can't register devAvme9440 at A16:%4.4x\n", BASEADD);
    return -1;
  }

  /* We end up here 1 time before all records are initialized */
  for (cardNum=0; cardNum < avme9440_num_links; cardNum++)
  {
    if (devAvme9440Debug >= 5)
    printf("devAvme9440: Init card %d\n", cardNum);

    probeVal = LED_INIT;  /* sends init value to LEDs and disables interrupts */

    if (devWriteProbe(1, &p->boardStatus, &probeVal))
    {

      if (devAvme9440Debug >= 5)
        printf("devAvme9440: Probe at %p failed\n", &p->boardStatus);
      cards[cardNum].card = NULL;		/* No card found */
    }
    else
    {
      if (devAvme9440Debug >= 5)
        printf("devAvme9440: Probe at %p success\n", &p->boardStatus);

      probeVal = INT_VEC_BASE + cardNum; 

      for (chanNum=0; chanNum <= 7; chanNum++)
      {
        p->iv[chanNum].vector = probeVal;
        if (devAvme9440Debug >= 5)
          printf("devAvme9440: Interrupt vector 0x%2.2X written to channel %d interrupt\n", probeVal, chanNum);

        cards[cardNum].intCnt[chanNum] = 0;  /* clear interrupt counter */
        scanIoInit(&cards[cardNum].ioscanpvt[chanNum]);  /* interrupt initialized per channel */
      }

      p->intEnable = 0x00;

      if (devAvme9440Debug >= 5)
        printf("devAvme9440: Interrupts disabled for channels 0 - 7\n");

      p->intTypeSelect = 0xFF;

      if (devAvme9440Debug >= 5)
      {
        printf("devAvme9440: Change of state interrupts set for channels 0 - 7\n");
        printf("             Interrupt level should be set at %d\n", int_level);
      }

      p->intPaternEnable = 0x00;

      if (devAvme9440Debug >= 5)
        printf("devAvme9440: Input pattern detection interrupts disabled for channels 0 - 7\n");

      p->boardStatus = LED_OKRUN;  /* sends init value to LEDs and disables interrupts */

      if (devAvme9440Debug >= 5)
        printf("devAvme9440: Card %d initialization complete\n", cardNum);

      cards[cardNum].card = p;		/* Remember address of the board */

      epicsAtExit(shutdown, (void *) p);

      if (devConnectInterruptVME(INT_VEC_BASE + cardNum, avme9440_isr,
              &cards[cardNum])) 
        printf("devAvme9440: Interrupt connect failed for card %d\n", cardNum);

      cards[cardNum].lock = epicsMutexMustCreate();
    }
    p++;
  }
  return 0;
}


/**************************************************************************
 *
 * Interrupt service routine
 *
 **************************************************************************/
static void avme9440_isr(void *parm)
{
   struct ioCard *pc = (struct ioCard *) parm;
   volatile struct avme9440 * p = pc->card;
   unsigned char intStatusLocal = p->intStatus;
   unsigned int chanNum;
   unsigned char   intStatWrite = 0;

   for (chanNum=0; chanNum <= 7; chanNum++)
   {
      if (intStatusLocal & (1 << chanNum))
      {
         scanIoRequest(pc -> ioscanpvt[chanNum]);
         intStatWrite |= (1 << chanNum);
      }
   }

   p->intStatus = intStatWrite;
}


/**************************************************************************
 *
 * BO Initialization (Called one time for each BO AVME9440 card record)
 *
 **************************************************************************/
static long init_bo_record(struct dbCommon *prec)
{
    struct boRecord *pbo = (struct boRecord *) prec;
    int status = 0;
    epicsUInt16 rval;
 
    /* bo.out must be an VME_IO */
    switch (pbo->out.type) {

    case VME_IO:

        if (pbo->out.value.vmeio.signal > 15)
        {
           pbo->pact = 1;          /* make sure we don't process this thing */
           status = S_db_badField;

           if (devAvme9440Debug >= 10)
              printf("devAvme9440: Illegal SIGNAL field ->%s<- \n", pbo->name);

           recGblRecordError(status,(void *)pbo,
            "devAvme9440 (init_record) Illegal SIGNAL field");
        }
        else
        {

           pbo->mask = 1;
           pbo->mask <<= pbo->out.value.vmeio.signal;

           if (read_card(pbo->out.value.vmeio.card, pbo->mask, &rval, OUTPUT_REG) == 0)
              {
              pbo->rbv = pbo->rval = rval;
              if (devAvme9440Debug >= 10)
                 printf("devAvme9440: ->%s<- Record initialized \n", pbo->name);
              }

           else 
              {
              status = 2;
              if (devAvme9440Debug >= 10)
                 printf("devAvme9440: ->%s<- Record failed to initialize \n", pbo->name);
              }

        }
        break;
         
    default :
        pbo->pact = 1;		/* make sure we don't process this thing */
        status = S_db_badField;

        if (devAvme9440Debug >= 10)
           printf("devAvme9440: Illegal OUT field ->%s<- \n", pbo->name);

        recGblRecordError(status,(void *)pbo,
            "devAvme9440 (init_record) Illegal OUT field");
    }
    return status;
}

/**************************************************************************
 *
 * BI Initialization (Called one time for each BI AVME9440 card record)
 *
 **************************************************************************/
static long init_bi_record(struct dbCommon *prec)
{
  struct biRecord *pbi = (struct biRecord *) prec;
  int status = 0;
  epicsUInt16 rval;

  switch (pbi->inp.type) {
  case VME_IO:

     if (pbi->inp.value.vmeio.signal > 31)
     {
        pbi->pact = 1;          /* make sure we don't process this thing */
        status = S_db_badField;

        if (devAvme9440Debug >= 10)
           printf("devAvme9440: Illegal SIGNAL field ->%s<- \n", pbi->name);

        recGblRecordError(status,(void *)pbi,
         "devAvme9440 (init_record) Illegal SIGNAL field");

     }
     else
     {
        pbi->mask = 1;
        if (pbi->inp.value.vmeio.signal <= 15)
           pbi->mask <<= pbi->inp.value.vmeio.signal;
        else
           pbi->mask <<= pbi->inp.value.vmeio.signal - 16;
        if (read_card(pbi->inp.value.vmeio.card, pbi->mask, &rval, INPUT_REG) == 0)
        {
           pbi->rval = rval;
           if (devAvme9440Debug >= 10)
              printf("devAvme9440: ->%s<- Record initialized \n", pbi->name);
        }

        else 
           {
           status = 2;
           if (devAvme9440Debug >= 10)
              printf("devAvme9440: ->%s<- Record failed to initialize \n", pbi->name);
           }
     }

     break;
  default:
     pbi->pact = 1;		/* make sure we don't process this thing */
     status = S_db_badField;
     if (devAvme9440Debug >= 10)
        printf("devAvme9440: Illegal INP field ->%s<- \n", pbi->name);
     recGblRecordError(status,(void *)pbi,
        "devAvme9440 (init_record) Illegal INP field");
   }
   return status;
}


/**************************************************************************
 *
 * MBBO Initialization (Called one time for each MBBO AVME9440 card record)
 *
 **************************************************************************/
static long init_mbbo_record(struct dbCommon *prec)
{
    struct mbboRecord *pmbbo = (struct mbboRecord *) prec;
    epicsUInt16 rval;
    struct vmeio *pvmeio;
    int status = 0;

    /* mbbo.out must be an VME_IO */
    switch (pmbbo->out.type) {
    case (VME_IO) :

        if (pmbbo->out.value.vmeio.signal > 15)
        {
           pmbbo->pact = 1;          /* make sure we don't process this thing */
           status = S_db_badField;

           if (devAvme9440Debug >= 10)
              printf("devAvme9440: Illegal SIGNAL field ->%s<- \n", pmbbo->name);

           recGblRecordError(status,(void *)pmbbo,
            "devAvme9440 (init_record) Illegal SIGNAL field");
        }
        else
        {

           pvmeio = &(pmbbo->out.value.vmeio);
           pmbbo->shft = pvmeio->signal;
           pmbbo->mask <<= pmbbo->shft;
 
           if (read_card(pmbbo->out.value.vmeio.card, pmbbo->mask, &rval, OUTPUT_REG) == 0) 
              {
              pmbbo->rbv = pmbbo->rval = rval;
              if (devAvme9440Debug >= 10)
                 printf("devAvme9440: ->%s<- Record initialized \n", pmbbo->name);
              }

           else
              {
              status = 2;
              if (devAvme9440Debug >= 10)
                 printf("devAvme9440: ->%s<- Record failed to initialize \n", pmbbo->name);
              }

        }
        break;

    default :

        pmbbo->pact = 1;          /* make sure we don't process this thing */
        status = S_db_badField;

        if (devAvme9440Debug >= 10)
           printf("devAvme9440: Illegal OUT field ->%s<- \n", pmbbo->name);

        recGblRecordError(status,(void *)pmbbo,
            "devAvme9440 (init_record) Illegal OUT field");
    }
    return status;
}

/**************************************************************************
 *
 * MBBI Initialization (Called one time for each MBBI AVME9440 card record)
 *
 **************************************************************************/
static long init_mbbi_record(struct dbCommon *prec)
{
    struct mbbiRecord *pmbbi = (struct mbbiRecord *) prec;
    int status = 0;
    epicsUInt16 rval;

    /* mbbi.inp must be an VME_IO */
    switch (pmbbi->inp.type) {
    case (VME_IO) :

     if (pmbbi->inp.value.vmeio.signal > 31)
     {
        pmbbi->pact = 1;          /* make sure we don't process this thing */
        status = S_db_badField;

        if (devAvme9440Debug >= 10)
           printf("devAvme9440: Illegal SIGNAL field ->%s<- \n", pmbbi->name);

        recGblRecordError(status,(void *)pmbbi,
         "devAvme9440 (init_record) Illegal SIGNAL field");

     }
     else
     {
        if (pmbbi->inp.value.vmeio.signal <= 15)
           pmbbi->shft = pmbbi->inp.value.vmeio.signal;
        else
           pmbbi->shft = pmbbi->inp.value.vmeio.signal - 16;
        pmbbi->mask <<= pmbbi->shft;
        if (read_card(pmbbi->inp.value.vmeio.card, pmbbi->mask, &rval, INPUT_REG) == 0)
        {
           pmbbi->rval = rval;
           if (devAvme9440Debug >= 10)
              printf("devAvme9440: ->%s<- Record initialized \n", pmbbi->name);
        }
        else
        {
           status = 2;
           if (devAvme9440Debug >= 10)
              printf("devAvme9440: ->%s<- Record failed to initialize \n", pmbbi->name);
        }
     }

     break;

     default :
        pmbbi->pact = 1;              /* make sure we don't process this thing */
        status = S_db_badField;
        if (devAvme9440Debug >= 10)
           printf("devAvme9440: Illegal INP field ->%s<- \n", pmbbi->name);
        recGblRecordError(status,(void *)pmbbi,
                "devAvme9440 (init_record) Illegal INP field");
        return status;
    }
    return 0;
}


/**************************************************************************
 *
 * Perform a write operation from a BO record
 *
 **************************************************************************/
static long write_bo(struct boRecord *pbo)
{

  epicsUInt16 rval;

  if (write_card(pbo->out.value.vmeio.card, pbo->mask, pbo->rval) == 0)
  {
    if(read_card(pbo->out.value.vmeio.card, pbo->mask, &rval, OUTPUT_REG) == 0)
    {
      pbo->rbv = rval;
      return 0;
    }
  }

  /* Set an alarm for the record */
  recGblSetSevr(pbo, WRITE_ALARM, INVALID_ALARM);
  return 2;
}

/**************************************************************************
 *
 * Perform a read operation from a BI record
 *
 **************************************************************************/
static long read_bi(struct biRecord *pbi)
{

  epicsUInt16 rval;
  int reg;

  if (pbi->inp.value.vmeio.signal <= 15)
     reg = INPUT_REG;
  else
     reg = OUTPUT_REG;

  if (read_card(pbi->inp.value.vmeio.card, pbi->mask, &rval, reg) == 0)
  {
     pbi->rval = rval;
     return 0;
  }

  /* Set an alarm for the record */
  recGblSetSevr(pbi, READ_ALARM, INVALID_ALARM);
  return 2;
}


/**************************************************************************
 *
 * Perform a write operation from a MBBO record
 *
 **************************************************************************/
static long write_mbbo(struct mbboRecord *pmbbo)
{

  epicsUInt16        rval;

  if (write_mbbo_card(pmbbo->out.value.vmeio.card, pmbbo->mask, pmbbo->rval) == 0)
  {
    if(read_card(pmbbo->out.value.vmeio.card, pmbbo->mask, &rval, OUTPUT_REG) == 0)
    {
      pmbbo->rbv = rval;
      return 0;
    }
  }

  /* Set an alarm for the record */
  recGblSetSevr(pmbbo, WRITE_ALARM, INVALID_ALARM);
  return 2;
}

/**************************************************************************
 *
 * Perform a read operation from a MBBI record
 *
 **************************************************************************/
static long read_mbbi(struct mbbiRecord *pmbbi)
{

  epicsUInt16        rval;
  int          reg;

  if (pmbbi->inp.value.vmeio.signal <= 15)
     reg = INPUT_REG;
  else
     reg = OUTPUT_REG;

  if (read_card(pmbbi->inp.value.vmeio.card, pmbbi->mask, &rval, reg) == 0)
  {
    pmbbi->rval = rval;
    return 0;
  }
  /* Set an alarm for the record */
  recGblSetSevr(pmbbi, READ_ALARM, INVALID_ALARM);
  return 2;
}


/**************************************************************************
 *
 * Raw read a bitfield from the card
 *
 **************************************************************************/
static long read_card(int card, epicsUInt32 mask, epicsUInt16 *prval, int reg)
{
  if (checkLink(card))
    return -1;

  if (reg == INPUT_REG)
    *prval = cards[card].card->inputData & mask;
  else
    *prval = cards[card].card->outputData & mask;

  if (devAvme9440Debug >= 20)
    printf("devAvme9440: read 0x%4.4X from card %d\n", *prval, card);

  return 0;
}

/**************************************************************************
 *
 * Raw write a bitfield to the card
 *
 **************************************************************************/
static long write_card(int card, epicsUInt32 mask, epicsUInt32 value)
{
  if (checkLink(card))
    return -1;

  epicsMutexMustLock(cards[card].lock);
  cards[card].card->outputData = (cards[card].card->outputData & ~mask) | value;
  epicsMutexUnlock(cards[card].lock);

  if (devAvme9440Debug >= 15)
    printf("devAvme9440: wrote 0x%4.4X to card %d\n", cards[card].card->outputData, card);

  return 0;
}

/**************************************************************************
 *
 * Raw write a bitfield to the card for MBBO records
 *
 **************************************************************************/
static long write_mbbo_card(int card, epicsUInt32 mask, epicsUInt32 value)
{
  if (checkLink(card))
    return -1;

  epicsMutexMustLock(cards[card].lock);
  cards[card].card->outputData = ((cards[card].card->outputData & ~mask) | (value & mask));
  epicsMutexUnlock(cards[card].lock);

  if (devAvme9440Debug >= 15)
    printf("devAvme9440: wrote 0x%4.4X to card %d\n", cards[card].card->outputData, card);

  return 0;
}


/**************************************************************************
 *
 * Make sure card number is valid
 *
 **************************************************************************/
static int checkLink(int card)
{
  if (card >= avme9440_num_links)
    return -1;
  if (cards[card].card == NULL)
    return -1;

  return 0;
}

#define ADDING   0
#define DELETING 1
/**************************************************************************
 *
 * BI record interrupt routine
 *
 **************************************************************************/
static long get_bi_int_info(int cmd, struct dbCommon *prec, IOSCANPVT *ppvt)
{
   struct biRecord *pbi = (struct biRecord *) prec;
   struct vmeio *pvmeio = (struct vmeio *)(&pbi->inp.value);
   volatile struct avme9440 *pc = cards[pvmeio->card].card;

   *ppvt = cards[pvmeio->card].ioscanpvt[pvmeio->signal];

   if (cmd == ADDING) /* enable interrupts */
   {
      devEnableInterruptLevelVME(int_level);

      pc->boardStatus = LED_OKINTS;  /* Initialize global board interrupts */

      pc->intEnable |= (1 << pvmeio->signal);  /* Initialize individual channel interrupts */

      cards[pvmeio->card].intCnt[pvmeio->signal]++;

   }
   else
      if(!(--(cards[pvmeio->card].intCnt[pvmeio->signal])))
         pc->intEnable &= ~(1 << pvmeio->signal);  /* Disable individual channel interrupts */

   return 0;
}

